package org.xbib.multiserv.plugin.tanserv;

import java.util.Random;

public class RandomTAN {

    private static final Random random = new MTRandom(false);

    private RandomTAN() {
    }

    public static String createPseudoTAN(int maxlen, boolean firstisletter, boolean luhn) {
        int digits = 8; // without luhn digit
        long l = random.nextLong();
        byte[] b = new byte[digits];
        for (int i = 0; i < digits; ++i) {
            b[i] = (byte) (l >> (digits - i - 1 << 3));
        }
        digits = Math.min(maxlen, 8);
        char[] ch = new char[digits + (luhn ? 1 : 0)];
        int pos = 0;
        if (firstisletter) {
            int b1 = b[pos] & 0xf; // 4 bit (0-15)
            int b2 = (b[pos] >> 4) & 0x3; // 3 bit (0-7)
            int b3 = (b[pos] >> 7) & 0x1; // 1 bit (0-1)
            ch[pos] = (char) (b1 + b2 + b3 + 65);  // 0..23
            pos = 1;
        }
        for (; pos < digits; pos++) {
            int lsb = b[pos] & 0x1f; // 5 bit (0-31)
            int msb = (b[pos] >> 5) & 0x3; // 2 bit (0-4)
            int d = 1 + lsb + msb; // random between 0 .. 35 (without 0, to avoid conflict with O)
            ch[pos] = codepoint2char(d);
        }
        if (luhn) {
            int sum = 0;
            int n = 35;
            int factor = 2;
            for (int i = 0; i < digits - 1; i++) {
                int offset = digits - 1 - i;
                char c = ch[offset];
                int d = char2codepoint(c);
                int weight = d * (factor - (i % factor));
                sum += ((weight < n) ? weight : (weight - (n -1)));
            }
            int p = (n - sum % n) % n;
            ch[pos] = codepoint2char(p);
            pos++;
        }
        return new String(ch, 0, pos);
    }

    public static boolean validateTAN(String s) {
        return validateTAN(s.toCharArray());
    }

    public static boolean validateTAN(char[] ch) {
        int sum = 0;
        int n = 35;
        int factor = 2;
        int len = ch != null ? ch.length : 0;
        if (len < 2 || len > 100) {
            return false;
        }
        for (int i = 0; i < len - 2; i++) {
            int offset = len - 2 - i;
            char c = ch[offset];
            int d = char2codepoint(c);
            if (d == 0) {
                // invalid character
                return false;
            } else {
                int weight = d * (factor - (i % factor));
                sum += ((weight < n) ? weight : (weight - (n - 1)));
            }
        }
        int p = (n - sum % n) % n;
        char character = codepoint2char(p);
        if (character == 0) {
            // invalid character
            return false;
        }
        return ch[len-1] == character;
    }

    private static char codepoint2char(int n) {
        return (char) ((n >= 0 && n <= 9) ? (n + 48) : (n >= 10 && n <= 35) ? (n + 55) : 0);
    }

    private static int char2codepoint(char n) {
        return (n >= 48 && n <= 57) ? n - 48 : (n >= 65 && n <= 90) ? n - 55 : 0;
    }
}
